# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.kucoin import kucoin


class kucoin2 (kucoin):

    def describe(self):
        # KuCoin v1 is deprecated, 'kucoin2' renamed to 'kucoin', 'kucoin2' to be removed on 2019-03-30
        return self.deep_extend(super(kucoin2, self).describe(), {
            'id': 'kucoin2',
        })
