# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.acx import acx


class bitibu (acx):

    def describe(self):
        return self.deep_extend(super(bitibu, self).describe(), {
            'id': 'bitibu',
            'name': 'Bitibu',
            'countries': ['CY'],
            'has': {
                'CORS': False,
            },
            'urls': {
                'logo': 'https://user-images.githubusercontent.com/1294454/45444675-c9ce6680-b6d0-11e8-95ab-3e749a940de1.jpg',
                'extension': '.json',
                'api': 'https://bitibu.com',
                'www': 'https://bitibu.com',
                'doc': 'https://bitibu.com/documents/api_v2',
            },
        })
